package com.boldchat.integrations;

import java.io.UnsupportedEncodingException;
import java.net.URLEncoder;
import java.nio.charset.Charset;
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import javax.servlet.http.HttpServletRequest;

public class ApiSecurity {
	private static final char[] HEX_CHARS = new char[] {
			'0', '1', '2', '3', '4', '5', '6', '7', '8', '9', 'A', 'B', 'C', 'D', 'E', 'F'
	};
	protected static final long DEFAULT_TIMEOUT = 300 * 1000;
	private final ApiSettingsFactory settingsFactory;
	private final String additionalPrivateKey;
	private String realAuth;
	private String realSignature;

	public ApiSecurity(ApiSettingsFactory settingsFactory) {
		this(settingsFactory, null);
	}

	public ApiSecurity(ApiSettingsFactory settingsFactory, String additionalPrivateKey) {
		this.settingsFactory = settingsFactory;
		this.additionalPrivateKey = additionalPrivateKey;
	}

	public ApiSettings verifyRequest(Long accountId, String url, HttpServletRequest request) throws AuthSecurityException {
		String authFromUser = request.getParameter("auth");
		String signatureFromUser = request.getParameter("signature");
		Map requestParams = request.getParameterMap();
		String requestUrl = forceSSL(url);
		return verifyRequest(accountId, authFromUser, signatureFromUser, requestUrl, requestParams);
	}

	public ApiSettings verifyRequest(Long accountId, String authFromUser, String signatureFromUser, String requestUrl, Map requestParams) throws AuthSecurityException {
		if(accountId == null || authFromUser == null || authFromUser.isEmpty()) {
			throw new AuthSecurityException("Not authorized");
		}

		AuthToken authToken = new AuthToken(authFromUser);

		if(!accountId.equals(authToken.getAccountId())) {
			throw new AuthSecurityException("AccountID does not match");
		}

		Long integrationId = authToken.getIntegrationId();

		ApiSettings apiSettings = settingsFactory.getIntegrationSettings(accountId, integrationId);

		if(apiSettings == null || !apiSettings.isEnabled()) {
			throw new AuthSecurityException("No integration settings");
		}

		Long timestamp = authToken.getTimestamp();
		checkTimeout(timestamp, apiSettings);

		realAuth = authToken.hasAuthHash() ? getAPIKey(accountId, integrationId, timestamp, getPrivateKey(apiSettings)) : null;
		realSignature = getAPISignature(getPrivateKey(apiSettings), requestUrl, requestParams);

		Long realAccountID = apiSettings.getAccountId();
		if(!accountId.equals(realAccountID) || hasInvalidAuthOrSignature(authFromUser, signatureFromUser)) {
			throw new AuthSecurityException("Invalid auth or signature");
		}

		return apiSettings;
	}

	public String getAPIKey(long accountId, long integrationId) throws AuthSecurityException {
		ApiSettings apiSettings = settingsFactory.getIntegrationSettings(accountId, integrationId);
		long timestamp = System.currentTimeMillis() + (5 * 60 * 1000);
		return getAPIKey(accountId, apiSettings.getSetupItemId(), timestamp, apiSettings.getPrivateKey());
	}

	public static String getAPIKey(Long accountID, Long setupItemID, Long timestamp, String privateKey) {
		String key = accountID + ":" + setupItemID + ":" + timestamp;
		if(privateKey == null) {
			return key;
		}
		String hash = getEncrypted(key, privateKey);
		return key + ":" + hash;
	}

	public static String getAPISignature(String privateKey, String requestUrl, Map requestParameters) {
		String unencryptedSignature = getUnencryptedSignature(requestUrl, requestParameters);
		return getEncrypted(unencryptedSignature, privateKey);
	}

	protected boolean hasInvalidAuthOrSignature(String authFromUser, String signatureFromUser) {
		return (!hasValueAndEqual(realAuth, authFromUser) && !hasValueAndEqual(realSignature, signatureFromUser)) || hasValueAndNotEqual(realSignature, signatureFromUser);
	}

	protected String getPrivateKey(ApiSettings apiSettings) {
		if(additionalPrivateKey == null) {
			return apiSettings.getPrivateKey();
		} else {
			return apiSettings.getPrivateKey() + ":" + additionalPrivateKey;
		}
	}

	protected static String forceSSL(String url) {
		return url.replace("http://", "https://");
	}

	protected static void checkTimeout(long timestamp, ApiSettings apiSettings) throws AuthSecurityException {
		Long timeout = apiSettings.getTimeoutMilliseconds();
		if(timeout == null) {
			timeout = DEFAULT_TIMEOUT;
		}
		if(System.currentTimeMillis() - timeout > timestamp) {
			throw new AuthSecurityException("Expired authorization");
		}
	}

	protected static boolean hasValueAndEqual(String realValue, String valueFromUser) {
		return realValue != null && realValue.equalsIgnoreCase(valueFromUser);
	}

	protected static boolean hasValueAndNotEqual(String realValue, String valueFromUser) {
		return realValue != null && !realValue.equalsIgnoreCase(valueFromUser);
	}

	protected static String getUnencryptedSignature(String requestUrl, Map requestParameters) {
		String signed = StringUtil.getAsString(requestParameters.get("signed"));
		if(signed == null || "".equals(signed)) {
			return null;
		}

		StringBuilder sb = new StringBuilder();
		sb.append(requestUrl);

		char separator = requestUrl.indexOf('?') >= 0 ? '&' : '?';
		String[] parts = signed.split("&");
		Set<String> setOfSignedParameters = new HashSet<>();
		for(String key : parts) {
			setOfSignedParameters.add(key);

			String val = StringUtil.getAsString(requestParameters.get(key));
			sb.append(separator);
			separator = '&';
			sb.append(urlEncode(key)).append("=").append(val == null ? "" : urlEncode(val));
		}

		if(!hasRequiredSignatureParams(setOfSignedParameters)) {
			return null;
		}

		return sb.toString();
	}

	protected static boolean hasRequiredSignatureParams(Set<String> setOfSignedParameters) {
		return setOfSignedParameters.contains("auth") && setOfSignedParameters.contains("signed");
	}

	private static String urlEncode(String value) {
		try {
			return URLEncoder.encode(value, "UTF-8");
		} catch(UnsupportedEncodingException e) {
			throw new AssertionError(e);
		}
	}

	/**
	 *
	 * @param unencrypted
	 * @param privateKey
	 * @return
	 */
	protected static String getEncrypted(String unencrypted, String privateKey) {
		if(unencrypted == null) {
			return null;
		}
		MessageDigest sha512 = getSha512();
		byte[] encrypted = sha512.digest((unencrypted + privateKey).getBytes(Charset.forName("UTF-8")));
		return hexString(encrypted);
	}

	private static String hexString(byte[] bytes) {
		StringBuilder sb = new StringBuilder();
		for(int i = 0; i < bytes.length; i++) {
			sb.append(HEX_CHARS[0x0F & (bytes[i] >> 4)]);
			sb.append(HEX_CHARS[0x0F & bytes[i]]);
		}
		return sb.toString();
	}

	private static MessageDigest getSha512() {
		try {
			return MessageDigest.getInstance("SHA-512");
		} catch(NoSuchAlgorithmException e) {
			throw new AssertionError(e);
		}
	}

	protected void setRealAuth(String realAuth) {
		this.realAuth = realAuth;
	}

	protected void setRealSignature(String realSignature) {
		this.realSignature = realSignature;
	}
}
