package com.boldchat.integrations;

import java.io.IOException;
import java.util.UUID;

import javax.servlet.Filter;
import javax.servlet.FilterChain;
import javax.servlet.FilterConfig;
import javax.servlet.ServletException;
import javax.servlet.ServletRequest;
import javax.servlet.ServletResponse;
import javax.servlet.http.Cookie;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

public abstract class ApiSecurityFilter implements Filter {
    private static final int UNAUTHORIZED = 401;
    private static final int SERVER_ERROR = 500;
	public static final String AUTH_PARAM_NAME = "auth";
	public static final String SIGNED_PARAM_NAME = "signed";
	public static final String SIGNATURE_PARAM_NAME = "signature";
	public static final String AUTH_ID_COOKIE_NAME = "_id";
    public static final String AUTH_COOKIE_NAME = "apiauth";
    public static final int AUTH_MAX_AGE_SECONDS = 24 * 60 * 60;

    private ApiSettingsFactory settingsFactory;

    public abstract Long getAccountId(HttpServletRequest request) throws ServletException;
    public abstract String getUrl(HttpServletRequest request) throws ServletException;

    @Override
    public void init(FilterConfig filterConfig) throws ServletException {
        String settingsFactoryName = filterConfig.getInitParameter("ApiSettingsFactory");
		settingsFactory = newSettingsFactory(settingsFactoryName);
    }

    protected static ApiSettingsFactory newSettingsFactory(String settingsFactoryName) throws ServletException {
        if(settingsFactoryName == null || settingsFactoryName.isEmpty()) {
            throw new ServletException("ApiSettingsFactory is a required parameter to ApiSecurityFilter");
        }
        try {
            Class factoryClass = Class.forName(settingsFactoryName);
            return (ApiSettingsFactory)factoryClass.newInstance();
        } catch (ClassNotFoundException | InstantiationException | IllegalAccessException e) {
            throw new ServletException(settingsFactoryName);
        }
    }

    @Override
    public void doFilter(ServletRequest servletRequest, ServletResponse servletResponse, FilterChain filterChain) throws ServletException, IOException {
        HttpServletRequest request = (HttpServletRequest) servletRequest;
        HttpServletResponse response = (HttpServletResponse) servletResponse;

        ApiSettings settings;
        String authId = ServletUtil.getCookieValue(request, AUTH_ID_COOKIE_NAME);
        try {
            if(hasAuthParam(request)) {
                ApiSecurity security = new ApiSecurity(settingsFactory);
                Long accountId = getAccountId(request);
                settings = security.verifyRequest(accountId, getUrl(request), request);
            } else {
                ApiSecurity security = new ApiSecurity(settingsFactory, authId);
                settings = verifyCookie(security, request);
            }
            if(settings == null) {
                deleteCookie(response);
                response.sendError(SERVER_ERROR);
                return;
            }
        } catch (AuthSecurityException e) {
            deleteCookie(response);
			response.sendError(UNAUTHORIZED);
			return;
        }

        authId = (authId == null || authId.isEmpty()) ? generateAuthId() : authId;
        setCookie(authId, settings, response);

        filterChain.doFilter(request, response);
    }

    protected void setCookie(String authId, ApiSettings settings, HttpServletResponse response) {
        String privateKey = settings.getPrivateKey() + ":" + authId;
        String apiAuth = ApiSecurity.getAPIKey(settings.getAccountId(), settings.getSetupItemId(), getCookieTimeout(), privateKey);

        Cookie cookie = new Cookie(AUTH_COOKIE_NAME, apiAuth);
        cookie.setMaxAge(AUTH_MAX_AGE_SECONDS);
        cookie.setSecure(true);
        response.addCookie(cookie);

		cookie = new Cookie(AUTH_ID_COOKIE_NAME, authId);
		cookie.setMaxAge(AUTH_MAX_AGE_SECONDS);
		cookie.setSecure(true);
		response.addCookie(cookie);
    }

    private String generateAuthId() {
        return UUID.randomUUID().toString();
    }

    protected long getCookieTimeout() {
        return System.currentTimeMillis() + (AUTH_MAX_AGE_SECONDS * 1000);
    }

    protected void deleteCookie(HttpServletResponse response) {
        Cookie cookie = new Cookie(AUTH_COOKIE_NAME, null);
        cookie.setMaxAge(0);
        cookie.setSecure(true);
        response.addCookie(cookie);
    }

    protected ApiSettings verifyCookie(ApiSecurity security, HttpServletRequest request) throws AuthSecurityException {
        String cookieValue = ServletUtil.getCookieValue(request, AUTH_COOKIE_NAME);
        if(cookieValue == null || cookieValue.isEmpty()) {
            throw new AuthSecurityException("No auth cookie");
        }

        AuthToken token = new AuthToken(cookieValue);
        String url;
        try {
            url = getUrl(request);
        } catch (ServletException e) {
            throw new AuthSecurityException("Failed to get URL", e);
        }
        url = ApiSecurity.forceSSL(url);
        return security.verifyRequest(token.getAccountId(), cookieValue, null, url, request.getParameterMap());
    }

    protected static boolean hasAuthParam(HttpServletRequest request) {
        String[] auth = request.getParameterValues(AUTH_PARAM_NAME);
        return auth != null && auth.length > 0;
    }



    @Override
    public void destroy() {
        settingsFactory = null;
    }
}
