package com.boldchat.integrations.sample;

import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import com.boldchat.integrations.ApiSecurity;
import com.boldchat.integrations.ApiSecurityFilter;
import com.boldchat.integrations.ApiSettings;
import com.boldchat.integrations.ApiSettingsFactory;
import com.boldchat.integrations.AuthSecurityException;
import com.boldchat.integrations.AuthToken;
import com.boldchat.integrations.BoldApiRequest;
import com.boldchat.integrations.sample.models.ReturnContext;
import com.boldchat.integrations.sample.models.TicketContext;
import com.google.gson.Gson;

public class SampleService {

    private Gson gson = new Gson();
    private static final String API_PATH_START = "/data/rest/json/v1/";
    private TicketContext[] allTickets;
    private final ApiSettingsFactory settingsFactory;

    public SampleService(ApiSettingsFactory settingsFactory) {
        this.settingsFactory = settingsFactory;
        this.allTickets = SampleData.getSampleData();
    }

    public String getTicket(String ticketID) {
        if (ticketID == null) {
            return this.buildErrorResponse("ticketID required");
        }

        int ticketIDint = 0;
         try {
            ticketIDint = Integer.parseInt(ticketID);
        } catch (Exception e) {
            return this.buildErrorResponse("ticketID must be numeric");
        }

        List<TicketContext> tickets = this.getTicketByTicketID(ticketIDint);

        if (tickets.size() == 0) {
            return this.buildErrorResponse("no results");
        } else if (tickets.size() > 1) {
            return this.buildErrorResponse("more than one ticket with ID " + ticketID);
        } else {
            return  this.buildSuccessResponse(tickets.get(0));
        }
    }

    public String searchTickets(String ticketID, String name, String email) {

        if (ticketID == null && email == null && name == null) {
            return this.buildErrorResponse("ticketID, name, or email required");
        }

        if (ticketID != null) {
            try {
                return this.checkHasResult(this.getTicketByTicketID(Integer.parseInt(ticketID)));
            } catch (Exception e) {
                return this.buildErrorResponse("ticketID must be numeric");
            }
        }

        if (email == null) {
            return this.checkHasResult(this.searchTicketsByName(name));
        }

        if (name == null) {
            return this.checkHasResult(this.searchTicketsByEmail(email));
        }

        List<TicketContext> combinedTicketResults = this.searchTicketsByName(name);
        combinedTicketResults.addAll(this.searchTicketsByEmail(email));
        combinedTicketResults = this.removeDuplicates(combinedTicketResults);

        return this.checkHasResult(combinedTicketResults);
    }

    public String linkTicket(Map<String, String[]> parameters, AuthToken authToken) {

        ReturnContext<String> result;

        try {
            result = gson.fromJson(this.performBoldRequest("updateChat", authToken, parameters), ReturnContext.class);
        } catch (IOException e) {
            return  this.buildErrorResponse(e.getMessage());
        } catch (AuthSecurityException e) {
            return  this.buildErrorResponse(e.getMessage());
        }

        if (result.getStatus().equals("error")) {
            return this.buildErrorResponse(result);
        }

        return this.buildSuccessResponse(result);
    }

    public String performBoldRequest(String apiMethod, AuthToken authToken, Map<String, String[]> parameters) throws IOException, AuthSecurityException {
        long accountId = authToken.getAccountId();
        long integrationId = authToken.getIntegrationId();
        ApiSettings apiSettings = settingsFactory.getIntegrationSettings(accountId, integrationId);
        BoldApiRequest request = new BoldApiRequest(apiSettings.getServerSet(), API_PATH_START + apiMethod, authToken.getAccountId());
        ApiSecurity security = new ApiSecurity(settingsFactory);
        request.addParameter("auth", security.getAPIKey(accountId, integrationId));
        Map<String, String[]> requestParams = removeAuthParameters(parameters);
        request.addAllParameters(requestParams);

        return request.fetchRawResult();
    }

    private List<TicketContext> getTicketByTicketID(int ticketID) {

        List<TicketContext> tickets = new ArrayList<>();

        for (TicketContext ticket : allTickets) {
            if (ticket.getTicketID() == ticketID) {
                tickets.add(ticket);
            }
        }

        return tickets;
    }

    private List<TicketContext> searchTicketsByName(String name) {

        List<TicketContext> tickets = new ArrayList<>();

        for (TicketContext ticket : allTickets) {
            if (ticket.getFirstName().toLowerCase().startsWith(name.toLowerCase()) ||
                    ticket.getLastName().toLowerCase().startsWith(name.toLowerCase()) ||
                    (ticket.getFirstName().toLowerCase() + ticket.getLastName().toLowerCase()).startsWith(name.toLowerCase())) {
                tickets.add(ticket);
            }
        }

        return tickets;
    }

    private List<TicketContext> searchTicketsByEmail(String email) {

        List<TicketContext> tickets = new ArrayList<>();

        for (TicketContext ticket : allTickets) {
            if (ticket.getEmail().toLowerCase().equals(email.toLowerCase())) {
                tickets.add(ticket);
            }
        }

        return tickets;
    }

    private String buildErrorResponse(String message){
        return gson.toJson(new ReturnContext(false, message));
    }

    private String buildErrorResponse(ReturnContext message) {
        return gson.toJson(new ReturnContext<>(false, message));
    }

    private String buildSuccessResponse(String message){
        return gson.toJson(new ReturnContext<>(true, message));
    }

    private String buildSuccessResponse(TicketContext tickets){
        return gson.toJson(new ReturnContext<>(true, tickets));
    }

    private String buildSuccessResponse(List tickets){
        return gson.toJson(new ReturnContext<>(true, tickets));
    }

    private String buildSuccessResponse(ReturnContext message) {
        return gson.toJson(new ReturnContext<>(true, message));
    }

    private List<TicketContext> removeDuplicates(List<TicketContext> list){
        for (int i = 1; i < list.size(); i++) {
            TicketContext a1 = list.get(i);
            TicketContext a2 = list.get(i-1);
            if (a1.equals(a2)) {
                list.remove(a1);
            }
        }

        return list;
    }

    private Map<String, String[]> removeAuthParameters(Map<String, String[]> params) {
        HashMap<String, String[]> newParams = new HashMap<>();
        for(String key : params.keySet()) {
            if(!key.equals(ApiSecurityFilter.AUTH_COOKIE_NAME)
                    && !key.equals(ApiSecurityFilter.AUTH_ID_COOKIE_NAME)
                    && !key.equals(ApiSecurityFilter.AUTH_PARAM_NAME)
                    && !key.equals(ApiSecurityFilter.SIGNATURE_PARAM_NAME)
                    && !key.equals(ApiSecurityFilter.SIGNED_PARAM_NAME)) {
                newParams.put(key, params.get(key));
            }
        }
        return newParams;
    }

    private String checkHasResult(List<TicketContext> tickets) {
        if (tickets.size() == 0) {
            return this.buildErrorResponse("no results");
        } else {
            return this.buildSuccessResponse(tickets);
        }
    }
}
